; Asteroid Belt Game - 24-Dec-2024 by Kevin Krausnick
; Inspired by articles in Issues 37 and 40 of Commodore Free Magazine (https://commodorefree.com/)
; Assembler: Acme
;
; memory map:
; 1k-2k   -> screen 0
; 2k-4k   -> ml program
; 8k-8.5k -> sprites
; 10k-11k -> screen 1
; 12k-16k -> custom character set
;
; sprite legend:
; 0 - space ship and thrusters (multicolor)
; 1 - not used
; 2 - scores (left 3 digits)
; 3 - scores (right 3 digits)
; 4 - fuel gauge (multicolor)
; 5 - level / lives
; 6 - lives shadow
; 7 - ufo

!addr zp_cap = $f9 ; zero byte address of space ship (capsule)
!addr zp_src = $fb ; zero byte source address
!addr zp_dst = $fd ; zero byte destination address
!addr sid = $d400  ; base address of sid registers
!addr joy = $dc00  ; joystick port 2 status
!addr row_btm_0 = $0400+40*24 ; base address of bottom row (screen 0)
!addr row_btm_1 = $2800+40*24 ; base address of bottom row (screen 1)
!addr row_color = $d800+40*24 ; base address of bottom row (color ram)

* = $801 ; basic header
!word line2:!pet $0a,$00,$8f,$20,"asteroid belt:24-dec-2024 by kpk",0:line2:!word prgend
!byte $14,$00,$9e,$20:!for i,3,0 {!pet '0'+start%10^(i+1)/10^i}:!byte 0:prgend:!byte 0,0

; ---------------------------------------------------------------------------------------

start
  jsr init      ; intialize program

select_skill    ; set skill level
  lda $d016     ; enable multicolor mode
  ora #$10
  sta $d016
  jsr show_skill_menu
- jsr $ffe4     ; wait for keypress
  beq -
  cmp #$03      ; run/stop key = exit game
  bne +
  jsr set_screen_0
  jsr game_end
  lda #$15      ; restore vic-ii settings
  sta $d018
  lda #$1b
  sta $d011     ; switch back to 25-row mode
  rts           ; exit game
+ cmp #$31      ; valid inputs are 1 - 5
  bcc -
  cmp #$36
  bcs -
  and #$0f
  tax
  dex
  lda skill_delays,x
  sta time_delay+1
  inx
  stx level_number
  lda #$03
  sta lives_count
  jsr reset_score
  lda #$93
  jsr $ffd2
  lda $d016     ; disable multicolor mode
  and #$ef
  sta $d016
  jsr main
  jmp select_skill

; ---------------------------------------------------------------------------------------

show_skill_menu   ; display skill level selection menu
  lda $d011       ; blank screen until rendered
  and #$ef
  sta $d011
  lda #>skill_level_msg
  sta zp_src+1
  lda #<skill_level_msg
  sta zp_src
  ldy #$00        ; show skill levels and stars
- lda (zp_src),y
  beq ssm_cont
  jsr $ffd2
  iny
  bne -
  inc zp_src+1
  bne -
ssm_cont
  lda #$d8        ; set color ram to multicolor for target range
  sta zp_dst+1
  lda #$ce
  sta zp_dst
  ldx #$0e
-- ldy #$13
  lda #$f9
- sta (zp_dst),y
  dey
  bpl -
  lda zp_dst
  clc
  adc #40
  sta zp_dst
  bcc +
  inc zp_dst+1
+ dex
  bne --
  lda $d011       ; reveal screen
  ora #$10
  sta $d011
  rts

; ---------------------------------------------------------------------------------------

main
  jsr game_start   ; start a new game
main_loop
  jsr fine_scroll
  jsr time_delay   ; control pace of program
  jsr animate_thrusters
  lda exit_flag
  bne main_exit
  lda game_over_flag
  bne main
  ldy capsule_x
  lda (zp_cap),y
  cmp #"."         ; check whether fuel pod has been intercepted
  bne +
  jsr gather_fuel
  jmp main_loop
+ cmp #$41         ; check for collision with asteroid
  bcc +
  jsr collision
  lda exit_flag
  beq main
  bne main_exit
+ jsr animate_ufo
  lda menu_flag
  bne main_exit
  lda exit_flag
  bne main
  jsr check_keyboard
  lda exit_flag
  beq main_loop
main_exit
  rts

; ---------------------------------------------------------------------------------------

check_keyboard
  lda $c6       ; check keyboard buffer
  beq ck_end
  jsr $ffe4
  cmp #$03      ; run/stop key = exit
  bne +
  jsr game_end
  lda #$01
  sta exit_flag
  rts
+ cmp #" "      ; space = pause
  bne +
  jsr game_pause
  jmp ck_end
+ cmp #"Q"      ; keyboard shortcut to level up
  bne +
  jsr level_up
  jmp ck_end
+ cmp #$91      ; cursor up to level up
  bne +
  jsr level_up
  jmp ck_end
+ cmp #"A"      ; keyboard shortcut to level down
  bne +
  jsr level_down
  jmp ck_end
+ cmp #$11      ; cursor down to level down
  bne +
  jsr level_down
  jmp ck_end
+ cmp #"Z"      ; keyboard shortcut to move left
  beq +
  cmp #$9d      ; cursor left
  beq +
  jmp ++
+ ldx capsule_x
  beq ck_end
  dec capsule_x
  jsr space_ship_move
  jmp ck_end
++ cmp #"M"     ; keyboard shortcut to move right
  beq +
  cmp #"X"      ; keyboard shortcut to move right
  beq +
  cmp #$1d      ; cursor right
  beq +
  jmp ck_end
+ ldx capsule_x
  cpx #$27
  beq ck_end
  inc capsule_x
  jsr space_ship_move
ck_end
+ rts

; ---------------------------------------------------------------------------------------

collision
  lda #$40
  sta (zp_cap),y      ; set collision char
  jsr sound_collision ; trigger sound
  lda $d015
  and #%11111110
  sta $d015           ; disable space ship sprite
  jmp game_over

gather_fuel
  lda #" "
  sta (zp_cap),y      ; erase the char
  jsr sound_add_fuel  ; trigger sound
  jmp inc_score       ; increment score

time_delay            ; control pace of program
  ldy #$00            ; immediate value is updated using self-modifying code
  ldx #$00
- dex
  bne -
  dey
  bne -
  rts

; ---------------------------------------------------------------------------------------

init
  lda #$93      ; initialize (clear) screen buffers
  jsr $ffd2
  jsr flip_screen
  lda #$93
  jsr $ffd2
  jsr flip_screen
  jsr set_screen_0
  jsr copy_charset
  jsr setup_custom_chars
  lda #$0b      ; dark gray
  sta $d020     ; border color
  lda #$00      ; black
  sta $d021     ; background color
  lda #$07      ; yellow
  sta $d022     ; background color 1
  lda #$09      ; brown
  sta $d023     ; background color 2
  lda #$08      ; orange
  sta $d025     ; sprite mc reg 0 (01 bit pattern)
  lda #$0a      ; light red
  sta $d026     ; sprite mc reg 1 (11 bit pattern)
  lda #$11
  sta $d01c     ; sprite 0 & 4 mc mode on
  lda #$80
  sta $28a      ; turn on key repeat
  lda #$08      ; lock upper case character set
  jsr $ffd2
  lda #$13
  sta $d011     ; select 24-row mode
  lda #$b8
  sta zp_cap    ; capsule location lo byte ($5b8 = screen 0, $29b8 = screen 1)
  lda #$00      ; clear sprite data
  tay
- sta $2000,y   ; data for sprites 0-3
  sta $2100,y   ; data for sprites 4-7
  iny
  bne -
  jsr show_lives_shadow
  lda #$01
  sta new_game
  jsr reset_score
  rts

; ---------------------------------------------------------------------------------------

copy_charset        ; copy default character set from ROM to RAM
  lda $dc0e
  and #$fe
  sta $dc0e         ; turn off keyscan interrupt timer
  lda $01
  and #$fb
  sta $01           ; switch in character
  lda #$00
  sta zp_src
  sta zp_dst        ; lo bytes for source and target addresses
  tay
  lda #$d0
  sta zp_src+1      ; hi byte source address
  lda #$30
  sta zp_dst+1      ; hi byte target address
  ldx #$10          ; copy 16 pages (4k) of character rom
- lda (zp_src),y
  sta (zp_dst),y
  iny
  bne -
  inc zp_src+1
  inc zp_dst+1
  dex
  bne -             ; copy character set
  lda $01
  ora #$04
  sta $01           ; switch in io
  lda $dc0e
  ora #$01
  sta $dc0e         ; restart keyscan interrupt timer
  rts

setup_custom_chars  ; overlay custom character set (two pages)
  ldy #$00
- lda alpha_chars,y ; a-z
  sta $3000+1*8,y   ; normal characters
  eor #$ff
  sta $3400+1*8,y   ; reversed characters
  iny
  cpy #$d0
  bne -
  ldy #$00
- lda numerals,y    ; 0-9
  sta $3000+48*8,y  ; normal characters
  eor #$ff
  sta $3400+48*8,y  ; reversed characters
  iny
  cpy #$50
  bne -
  ldx #$00
- lda custom_chars,x
  sta $3000+64*8,x  ; normal characters
  eor #$ff
  sta $3400+64*8,x  ; reversed characters
  inx
  cpx #23*8         ; 23 custom character bitmap objects
  bne -
  rts

; ---------------------------------------------------------------------------------------

fine_scroll
- lda $d012     ; wait for visible raster line
  cmp #$30
  bcs -
  lda frame_no+1
  cmp #$07
  bne +
  jsr fs_flip
+ lda $d011
  and #$f0
frame_no
  ora #$00
  sta $d011
  dec frame_no+1
  lda frame_no+1
  and #$07
  sta frame_no+1
  rts

fs_flip
  jsr flip_screen   ; switch between screen 0 and screen 1
  jsr random_number ; place new fuel pod in random column on bottom row
  lda #"."
  sta row_btm_0,x   ; place fuel pod on screen 0
  sta row_btm_1,x   ; place fuel pod on screen 1
  lda #$07          ; yellow
  sta row_color,x   ; colorize character
  jsr random_number ; place new asteroid in random column on bottom row
  txa
  and #$03          ; random asteroid
  clc
  adc #$41
  pha
  jsr random_number ; random x location
  pla
  sta row_btm_0,x   ; place asteroid on screen 0
  sta row_btm_1,x   ; place asteroid on screen 1
  lda #$0f          ; light gray
  sta row_color,x   ; colorize character
  dec fuel_qty      ; decrease amount of available fuel
  jsr update_fuel_gauge
  lda fuel_qty
  bne ++
  lda #$00          ; disable thrusters sprite (black)
  sta $d026         ; sprite mc reg 1 (11 bit pattern)
  ldx #$00
- lda out_of_fuel_msg,x
  beq +
  sta $526,x        ; display out-of-fuel message (screen 0)
  sta $2926,x       ; display out-of-fuel message (screen 1)
  lda #$01          ; white
  sta $d926,x       ; set text color
  inx
  bne -
+ jsr game_over     ; ran out of fuel, so game over
  lda #$00
  sta $c6           ; clear keyboard buffer
  lda #$01
  sta game_over_flag
++ jsr check_joystick
  rts

; ---------------------------------------------------------------------------------------

flip_screen     ; switch between screen 0 and screen 1
  jsr move_color_ram
  lda $288
  cmp #$04
  bne +
  jsr screen_0to1
  rts
+ jmp screen_1to0

screen_0to1     ; switch from screen 0 to screen 1 after copying data
  ldy #$00
- lda $4a0,y
  sta $2878,y
  lda $5a0,y
  sta $2978,y
  lda $6a0,y
  sta $2a78,y
  lda $6a0,y
  sta $2a78,y
  lda $710,y
  sta $2ae8,y
  iny
  bne -
  ldy #39
  lda #" "
- sta $2bc0,y
  dey
  bpl -
  jmp set_screen_1

screen_1to0     ; switch from screen 1 to screen 0 after copying data
  ldy #$00
- lda $28a0,y
  sta $478,y
  lda $29a0,y
  sta $578,y
  lda $2aa0,y
  sta $678,y
  lda $2b10,y
  sta $6e8,y
  iny
  bne -
  ldy #39
  lda #" "
- sta $7c0,y
  dey
  bpl -
  jmp set_screen_0

set_screen_0    ; set screen memory to 1k
- lda $d011
  bpl -
  lda #$1c
  sta $d018
  lda #$04
  sta $288
  lda #$05      ; space ship location hi byte ($5b8 = screen 0)
  sta zp_cap+1
  rts

set_screen_1    ; set screen memory to 10k
  lda #$ac
  sta $d018
  lda #$28
  sta $288
  lda #$29      ; space ship location hi byte ($29b8 = screen 1)
  sta zp_cap+1
  rts

move_color_ram  ; scroll color ram up one row, to match characters
  lda #$d8
  sta zp_src+1
  sta zp_dst+1
  lda #$a0
  sta zp_src
  lda #$78
  sta zp_dst
  ldx #21
-- ldy #$00
- lda (zp_src),y
  sta (zp_dst),y
  iny
  cpy #40
  bne -
  lda zp_src
  clc
  adc #40
  sta zp_src
  bcc +
  inc zp_src+1
+ lda zp_dst
  clc
  adc #40
  sta zp_dst
  bcc +
  inc zp_dst+1
+ dex
  bne --
  rts

; ---------------------------------------------------------------------------------------

animate_thrusters
  lda $a2
  sbc thrust_timer ; animate space ship thrusters
  bpl +
  eor #255
+ cmp #4
  bcc +
  lda $a2
  sta thrust_timer
  lda $d026     ; sprite mc reg 1 (11 bit pattern)
  eor #$0a      ; toggle between light red and black
  sta $d026
+ rts

; ---------------------------------------------------------------------------------------

animate_ufo
  ldx #$01
- inc $d00e        ; sprite 7 x pos
  bne +
  lda $d010        ; sprites x pos msb
  eor #%10000000
  sta $d010
+ dex
  bne -
  lda ufo_y_dir
  beq +
  inc $d00f        ; sprite 7 y pos
  bne ufo_cont
  jsr ufo_new_pos
  jmp ufo_cont
+ dec $d00f
  bne +
  jsr ufo_new_pos
  jmp ufo_cont
+ dec $d00f
  bne ufo_cont
  jsr ufo_new_pos
  jmp ufo_cont
ufo_cont
  lda $d01e        ; check for sprites collision
  and #%00000001
  beq +
  lda #$01
  sta exit_flag
  lda $d015
  and #%01111110
  sta $d015        ; disable ufo and space ship sprites
  lda #$0f         ; light gray
  ldy capsule_x
  sta $d800+40*11,y ; colorize character
  jmp collision
+ rts

ufo_new_pos
  lda sid+27       ; randomize ufo x pos
  sta $d00e        ; new sprite 7 x pos starting position
  ldx #$00         ; brief time delay
- dex
  bne -
  lda #$01
  sta animate_ufo+1
  lda sid+27       ; randomize ufo x velocity
  pha
  and #$01
  beq +
  inc animate_ufo+1
+ pla
  pha
  and #$02
  beq +
  lda animate_ufo+2
  eor #$20         ; toggle between inc / dec x pos above
  sta animate_ufo+2
+ ldx #$00
  pla
  and #$04
  beq +
  inx
+ stx ufo_y_dir
  rts

; ---------------------------------------------------------------------------------------

update_fuel_gauge
  lda #<fuel_gauge_sprite
  sta zp_src
  lda #>fuel_gauge_sprite
  sta zp_src+1
  lda fuel_qty    ; convert fuel qty to eight tiers of 12.5% per tier
  beq uf_main
  lsr
  lsr
  lsr
  lsr
  lsr
  and #$07
  tax
  inx
  cpx #$01
  beq +
  lda #$07        ; yellow
  sta $d02b       ; color sprite 4 (fuel gauge)
  jmp uf_cont
+ lda $a2
  sbc fuel_timer
  bpl +
  eor #$ff
+ cmp #$0f
  bcc uf_cont
  lda $a2
  sta fuel_timer
  lda $d02b       ; blink fuel gauge if critically low
  eor #$07
  sta $d02b       ; color sprite 4 (fuel gauge)
uf_cont
- lda zp_src
  clc
  adc #$03
  sta zp_src
  bcc +
  inc zp_src+1
+ dex
  bne -
uf_main
  ldy #$02        ; copy fuel qty pct to sprite 4 data location
- lda (zp_src),y  ; copy 5 rows of bitmapped data
  sta 132*64+3,y
  sta 132*64+6,y
  sta 132*64+9,y
  sta 132*64+12,y
  sta 132*64+15,y
  dey
  bpl -
  rts

; ---------------------------------------------------------------------------------------

game_end
  lda #$00
  sta sid+24    ; mute sound
  ldx #$00
  stx $d015     ; disable all sprites
  lda #$93      ; clear screen
  jsr $ffd2
  rts

; ---------------------------------------------------------------------------------------

game_start         ; initialize sprites
  lda #$00
  sta game_over_flag
  sta exit_flag
  sta menu_flag
  ldy capsule_x
  ldy #$13         ; default capsule column position
  sty capsule_x
  lda #$ff         ; start with full tank of fuel (255 units)
  sta fuel_qty
  ldy #$07         ; eight sprites (0-7)
  ldx #135         ; sprite base addresses (128 to 135 * 64 = $2000 to $21c0)
- txa
  sta $7f8,y       ; set sprite data pointers (screen 0)
  sta $7f8+$2400,y ; set sprite data pointers (screen 1)
  dex
  dey
  bpl -
  ldy #$07         ; eight sprites (0-7)
- lda sprite_colors,y
  sta $d027,y      ; set sprite colors
  dey
  bpl -
  ldy #$0f         ; set sprite x/y positions
- lda sprite_xy_pos,y
  sta $d000,y
  dey
  bpl -
  ldx #$00
  ldy #$00
- lda ufo_sprite,x
  sta 135*64,y     ; copy ufo sprite to data to sprite 7
  iny
  iny
  iny
  inx
  cpx #$08
  bne -
  ldy #$00
- lda space_ship_sprite,y
  sta 128*64,y   ; copy space ship sprite data to sprite 0
  iny
  cpy #$1c
  bne -
  jsr show_level_number
  lda #$55         ; set fuel qty pct top and bottom borders for sprite 4
  sta 132*64+0
  sta 132*64+1
  sta 132*64+18
  sta 132*64+19
  lda #$50
  sta 132*64+2
  sta 132*64+20
  lda $a2
  sta thrust_timer
  jsr get_ready
  lda exit_flag
  beq +
  rts
+ lda #%11111101
  sta $d015        ; enable all active sprites
  jsr init_sound
  jmp space_ship_move

; ---------------------------------------------------------------------------------------

get_ready
  jsr update_fuel_gauge
  lda #%11111100
  sta $d015     ; enable all active sprites except space ship
  jsr show_lives
  lda new_game
  beq +
  lda #$00
  sta new_game
  jsr reset_score
+ lda #$93      ; clear screen
  jsr $ffd2
  ldx #$00
- lda ready_msg,x
  beq +
  sta $577,x    ; display get ready message - line 2 (screen 0)
  sta $2977,x   ; display get ready message - line 2 (screen 1)
  lda #$01      ; white
  sta $d977,x   ; set text color
  inx
  bne -
+ lda $a2       ; wait for two seconds or keypress - whichever comes first
  clc
  adc #$78
  tax
- cpx $a2
  beq +
  jsr $ffe4
  beq -
+ cmp #$03      ; run/stop key = exit
  bne +
  lda #$01
  sta exit_flag
  jmp game_end
+ lda #$93      ; clear screen
  jsr $ffd2
  dec lives_count
  jmp show_lives

; ---------------------------------------------------------------------------------------

level_up
  lda level_number
  cmp #5
  beq +
  inc level_number
- ldx level_number
  dex
  lda skill_delays,x
  sta time_delay+1
  jsr show_level_number
+ rts

level_down
  lda level_number
  cmp #1
  beq +
  dec level_number
  bne -
+ rts

show_level_number
  ldx #$00
  ldy #$00
- lda $3060,x      ; copy letter 'l' from character set
  sta 133*64,y
  iny
  iny
  iny
  inx
  cpx #$08
  bne -
  lda level_number
  asl
  asl
  asl
  tax
  ldy #$01
- lda $3180,x
  sta 133*64,y     ; copy level number from character set
  iny
  iny
  iny
  inx
  cpy #$16
  bne -
  ldx #$00         ; shift right four pixels to center text above lives
  txa
-- sta 133*64+2,x
  ldy #$04
- lsr 133*64,x
  ror 133*64+1,x
  ror 133*64+2,x
  dey
  bne -
  inx
  inx
  inx
  cpx #$18
  bne --
  rts

; ---------------------------------------------------------------------------------------

show_lives
  ldy #$3e
  lda #$00      ; clear lives sprite target locations
- sta 133*64+24,y ; data for sprite 5 (second row)
  dey
  bpl -
  lda #$21
  sta zp_dst+1
  lda #$58
  sta zp_dst
  ldx lives_count
  beq +
  ldy #$00
- jsr show_cap  ; show 1, 2 or 3 capsules representing lives
  iny
  dex
  bne -
+ rts

show_lives_shadow
  lda #$21
  sta zp_dst+1
  lda #$98
  sta zp_dst
  ldy #$02
- jsr show_cap   ; show shadows for 3 capsules representing lives
  dey
  bpl -
  rts

show_cap
  txa
  pha
  tya
  pha
  ldx #$00
- lda space_ship_shadow,x
  sta (zp_dst),y ; copy lives sprite data
  iny
  iny
  iny
  inx
  cpx #$08
  bne -
  pla
  tay
  pla
  tax
  rts

; ---------------------------------------------------------------------------------------

init_sound
  ldx #$18      ; load sid registers with default values
- lda sid_defaults,x
  sta sid,x
  dex
  bpl -
  rts

sound_add_fuel
  lda #$20
  sta sid+1     ; v1 freq hi
  lda #$02
  sta sid+6     ; v1 sustain/release
  lda #$11      ; triangle wave on
  sta sid+4     ; v1 gate on
  ldx #$00      ; brief time delay to allow sound to play
- dex
  bne -
  lda #$10      ; triangle wave off
  sta sid+4     ; v1 gate off
  rts

sound_add_life
  lda #$80
  sta sid+1     ; v1 freq hi
  lda #$a8
  sta sid+6     ; v1 sustain/release
  lda #$41      ; pulse wave on
  sta sid+4     ; v1 gate on
  ldx #$00      ; brief time delay to allow sound to play
- dex
  bne -
  lda #$40      ; pulse wave off
  sta sid+4     ; v1 gate off
  rts

sound_collision
  lda #$80      ; noise wave off
  sta sid+11    ; v2 gate off
  lda #$19
  sta sid+12    ; v2 attack/decay
  lda #$09
  sta sid+13    ; v2 sustain/release
  lda #$20
  sta sid+22    ; filter freq hi
  lda #$81      ; noise wave on
  sta sid+11    ; v2 gate on
  ldx #$00      ; brief time delay to allow sound to play
- dex
  bne -
  lda #$80      ; noise wave off
  sta sid+11    ; v2 gate off
  rts

; ---------------------------------------------------------------------------------------

game_pause      ; pause game
  lda #$00
  sta sid+24    ; mute sound
  lda #$07      ; yellow
  sta $d02b     ; color sprite 4 (fuel gauge) (in case it's blinking off when paused)
  lda #$00
  sta $c6       ; clear keyboard buffer
- lda joy       ; check joystick fire button
  and #$10
  beq -         ; if pressed, wait until it is released
  lda #$08      ; orange
  sta pause_color
pause_loop
  lda $a2       ; get current jiffy clock value
  clc
  adc #$1e      ; half-second delay between flashes
  sta pause_timer
-- ldx #$00
- lda pause_msg,x
  beq +
  sta $529,x    ; display pause message (screen 0)
  sta $2929,x   ; display pause message (screen 1)
  lda pause_color
  sta $d929,x   ; set text color
  inx
  bne -
+ lda pause_timer
  cmp $a2       ; check to see if it's time to swap colors
  bne +
  lda pause_color
  eor #$08      ; toggle between black and orange
  sta pause_color
  jmp pause_loop
+ lda joy       ; check joystick fire button
  and #$10
  bne +
- lda joy       ; if fire button has been pressed...
  and #$10
  beq -         ; ...wait until it is released
  bne unpause
+ jsr $ffe4     ; check for keypress
  beq --
  cmp #$03      ; run/stop key = exit
  bne unpause
  lda #$01
  sta exit_flag
  jmp game_end
unpause
  ldx #$04
  lda #" "
- sta $529,x    ; erase pause message (screen 0)
  sta $2929,x   ; erase pause message (screen 1)
  dex
  bpl -
  lda #$1f
  sta sid+24    ; unmute sound
  rts

; ---------------------------------------------------------------------------------------

game_over
  lda #$13
  sta $d011     ; turn off fine scroll
  lda #$07      ; yellow
  sta $d02b     ; color sprite 4 (fuel gauge)
  lda #$80
  sta sid+11    ; turn off v2

  lda lives_count
  bne +
  lda #$03
  sta lives_count
  sta new_game
  ldx #$00
- lda game_over_msg,x
  beq +
  sta $577,x    ; display game over message (screen 0)
  sta $2977,x   ; display game over message (screen 1)
  lda #$01      ; white
  sta $d977,x   ; set text color
  inx
  bne -
+
  lda $a2       ; wait for half a second
  clc
  adc #$1e
- cmp $a2
  bne -
  lda #$00
  sta $c6       ; clear keyboard buffer
-- lda joy      ; check joystick fire button
  and #$10
  bne +
- lda joy       ; if fire button has been pressed...
  and #$10
  beq -         ; ...wait until it is released
  bne go_exit
+ jsr $ffe4
  beq --
  cmp #$03      ; run/stop key = exit
  bne go_exit
  lda #$01
  sta exit_flag
  sta menu_flag
  jsr game_end
go_exit
  rts

; ---------------------------------------------------------------------------------------

reset_score          ; set score to 0 ("000000")
  lda #$00
  sta score
  sta score+1
  sta bonus_count
  ldy #$05
- sta score_digits,y
  dey
  bpl -
  jmp show_score

inc_score            ; increment (non-weighted) score values
  inc score
  bne +
  inc score+1
+ inc bonus_count
  lda bonus_count    ; bonus life every 1000 points
  cmp #100           ; 100 * 10 (implied) = 1000
  bne +
  lda #0
  sta bonus_count
  lda lives_count
  cmp #3
  bcs +
  inc lives_count
  jsr show_lives
  jsr sound_add_life
+ ldx #$04           ; increment score by 10 - start with tens column
- inc score_digits,x ; increment by one
  lda score_digits,x
  cmp #$0a           ; check for overflow
  bcc +
  lda #$00
  sta score_digits,x ; on overflow, set to zero...
  dex                ; and increment column to the left
  bpl -
+ jsr show_score
  lda fuel_qty
  cmp #$f6           ; check whether fuel tanks are already "full"
  bcs +
  adc #$0a           ; there's room, so add 10 units of fuel
  sta fuel_qty
+ rts

show_score           ; show score using two adjacent sprites
  lda #<score_digits
  sta zp_src
  lda #>score_digits
  sta zp_src+1
  lda #$98           ; second row of sprite 2 data (score left 3 digits)
  sta zp_dst
  lda #$20
  sta zp_dst+1
  jsr copy_digits

  ldx score          ; replace high score if score is bigger
  cpx hi_score
  lda score+1
  sbc hi_score+1
  bcc +
  stx hi_score
  lda score+1
  sta hi_score+1
  jsr copy_high_score
+ jmp show_high_score

show_high_score      ; show high score using two adjacent sprites
  lda #<hi_score_digits
  sta zp_src
  lda #>hi_score_digits
  sta zp_src+1
  lda #$80           ; first row of sprite 2 data (high score left 3 digits)
  sta zp_dst
  lda #$20
  sta zp_dst+1
  jsr copy_digits
  rts

copy_high_score      ; copy score to high score
  ldy #$05
- lda score_digits,y
  sta hi_score_digits,y
  dey
  bpl -
  rts

copy_digits          ; copy group of six digits to two sprites
  ldy #$00
- lda (zp_src),y
  sta $02
  jsr copy_digit
  inc zp_dst
  iny
  cpy #$03
  bne +
  lda zp_dst
  adc #$3c
  sta zp_dst
+ cpy #$06
  bne -
  rts

copy_digit           ; copy digit character bitmap to sprite
  tya
  pha
  txa
  pha
  lda $02
  asl                ; multiply by 8 (eight bytes per character/digit)
  asl
  asl
  tax                ; make x the offset into character definitions
  ldy #$00           ; make y the offset into sprite data
- lda numerals,x
  sta (zp_dst),y
  inx                ; next byte (row) of character data
  iny                ; sprite is 24 bits wide, but we are only using 8 bits
  iny                ; ...so skip the other two bytes
  iny
  cpy #$18           ; repeat until the end of sprite is reached
  bne -
  pla
  tax
  pla
  tay
  rts

; ---------------------------------------------------------------------------------------

check_joystick
  lda joy       ; joystick (port 2)
  and #$01      ; up
  bne +
  jsr level_up
  jmp cj_end
+ lda joy
  and #$02      ; down
  bne +
  jsr level_down
  jmp cj_end
+ lda joy
  and #$04      ; left
  bne +
  ldx capsule_x
  beq cj_end
  dec capsule_x
  jsr space_ship_move
  jmp cj_end
+ lda joy
  and #$08      ; right
  bne +
  ldx capsule_x
  cpx #$27
  beq cj_end
  inc capsule_x
  jsr space_ship_move
+ lda joy
  and #$10      ; fire
  bne cj_end
  jsr game_pause
cj_end
  rts

; ---------------------------------------------------------------------------------------

space_ship_move ; align space ship sprite to character location
  lda capsule_x
  cmp #$1d
  bcc +
  sbc #$1d      ; x pos = 256-319
  asl           ; pixel = char * 8
  asl
  asl
  sta $d000     ; sprite 0 x pos
  lda $d010
  ora #%00000001
  sta $d010     ; sprite 0 x pos msb on
  rts
+ asl           ; x pos = 0-255
  asl           ; pixel = char * 8
  asl
  adc #$18      ; offset from left edge of screen
  sta $d000     ; sprite 0 x pos
  lda $d010
  and #%11111110
  sta $d010     ; sprite 0 x pos msb off
  rts

; ---------------------------------------------------------------------------------------

random_number   ; generate a random # between 0-39
- lda sid+27
  and #%00111111
  cmp #$28
  bcs -
  tax
  rts

; ---------------------------------------------------------------------------------------

sid_defaults
!byte $00,$20,$00,$01,$00,$00,$02 ; v1
!byte $00,$04,$00,$00,$81,$00,$c0 ; v2
!byte $00,$ff,$00,$00,$80,$00,$00 ; v3
!byte $00,$04,$02,$1f ; vol/filt

skill_delays    !byte 12,9,3,2,1
hi_score        !fill 2,0
hi_score_digits !fill 6,0
ready_msg       !scr "get ready",0
pause_msg       !scr "pause",0
out_of_fuel_msg !scr "out of fuel",0
game_over_msg   !scr "game over",0

skill_level_msg
!hex 059311111d1d53454c45435420534b494c4c204c4556454c3a0d1111
!hex 1d1dcd1d1d1dc5c6c7c80d
!hex 1d1dce1db71dc9cacbcc0d11
!hex 1d1dcf1d1d1dc5c6c7c8c5c6c7c80d
!hex 1d1dd01db71dc9cacbccc9cacbcc0d11
!hex 1d1dd11d1d1dc5c6c7c8c5c6c7c8c5c6c7c80d
!hex 1d1dd21db71dc9cacbccc9cacbccc9cacbcc0d11
!hex 1d1dd31d1d1dc5c6c7c8c5c6c7c8c5c6c7c8c5c6c7c80d
!hex 1d1dd41db71dc9cacbccc9cacbccc9cacbccc9cacbcc0d11
!hex 1d1dd51d1d1dc5c6c7c8c5c6c7c8c5c6c7c8c5c6c7c8c5c6c7c80d
!hex 1d1dd61db71dc9cacbccc9cacbccc9cacbccc9cacbccc9cacbcc00

alpha_chars ; a-z
!hex fec6c6fec6e6e600fcc6c6fce6e6fc00fec6c0c0e0e6fe00f8ccc6c6e6e6fe00
!hex fec0c0f8e0e0fe00fe00c0f8e0e0e000fec6c0cee6e6fe00c6c6c6fee6e6e600
!hex 7c30303038387c001e0c0c0c1cdc7800c6ccd8d0f8ece600c0c0c0c0e0e0fe00
!hex c6eefed6e6e6e600c6d6dedeeee6e600fec6c6e6e6e6fe00fec6c6fee0e0e000
!hex fec6c6c6eeeeff00fec6c6fef8ece600fec6c0fe06e6fe00fe30303038383800
!hex c6c6c6c6e6e6fe00c6c6c6e6e6fe7c00c6c6d6deeee6e600c6c67c387ce6e600
!hex c6c6c67c30383800fe060c3870e0fe00

numerals ; 0-9
!hex 7cc6ced6e6e67c00307070303838fe007ce6060c70e0fe007cc6061c06c67c00
!hex 1e3666c6ff060600fec0fc0606e6fc007cc6c0fce6e67c00fec60c3030383800
!hex 7cc6c67ce6e67c007cc6c67e06e67c00

custom_chars
!byte $10,$46,$34,$2a,$14,$5a,$24,$00 ; collision
!byte $38,$26,$e1,$87,$84,$42,$49,$36 ; asteroid 1
!byte $00,$18,$2c,$66,$42,$34,$18,$00 ; asteroid 2
!byte $00,$00,$18,$2c,$24,$18,$00,$00 ; asteroid 3
!byte $32,$55,$59,$82,$81,$e1,$26,$38 ; asteroid 4
!byte $00,$00,$00,$00,$00,$00,$15,$2a ; multicolor star
!byte $01,$01,$01,$01,$05,$05,$55,$55
!byte $00,$80,$80,$80,$a0,$a0,$a5,$95
!byte $00,$00,$00,$00,$00,$00,$54,$a8
!byte $02,$00,$00,$00,$00,$01,$02,$02
!byte $a9,$a5,$96,$5a,$6a,$a8,$80,$00
!byte $5a,$5a,$96,$a5,$a9,$2a,$02,$00
!byte $80,$00,$00,$00,$00,$40,$80,$80
double_height_chars
!byte $00,$0c,$1c,$3c,$3c,$0c,$0c,$0c ; 1
!byte $0c,$0c,$0c,$0c,$0c,$3f,$3f,$3f
!byte $00,$3e,$7f,$7f,$63,$63,$03,$06 ; 2
!byte $0c,$18,$30,$60,$60,$7f,$7f,$7f
!byte $00,$3e,$7f,$7f,$63,$63,$03,$0f ; 3
!byte $0e,$0f,$03,$63,$63,$7f,$7f,$3e
!byte $00,$63,$63,$63,$63,$63,$63,$7f ; 4
!byte $7f,$7f,$03,$03,$03,$03,$03,$03
!byte $00,$7f,$7f,$7f,$60,$60,$60,$7e ; 5
!byte $7f,$7f,$03,$63,$63,$7f,$7f,$3e

sprite_xy_pos
!byte $b8,$8f,$00,$00,$a0,$37,$b8,$37,$d8,$3b,$80,$37,$80,$37,$00,$00

sprite_colors
!byte $03,$0a,$05,$05,$07,$03,$0b,$0a

space_ship_sprite ; multicolor
!byte $c0,$c0,$00,$80,$80,$00,$a2,$80,$00,$aa,$80,$00,$aa,$80
!byte $00,$2a,$00,$00,$22,$00,$00,$2a,$00,$00,$08,$00,$00,$08

space_ship_shadow
!byte $41,$63,$7f,$7f,$3e,$14,$1c,$08

ufo_sprite
!byte $00,$00,$3c,$ff,$7e,$00,$00,$00

fuel_gauge_sprite ; multicolor
!byte %01000000,%00000000,%00010000 ; 0%
!byte %01100000,%00000000,%00010000 ; 13%
!byte %01101000,%00000000,%00010000 ; 25%
!byte %01101010,%00000000,%00010000 ; 38%
!byte %01101010,%10000000,%00010000 ; 50%
!byte %01101010,%10100000,%00010000 ; 63%
!byte %01101010,%10101000,%00010000 ; 75%
!byte %01101010,%10101010,%00010000 ; 88%
!byte %01101010,%10101010,%10010000 ; 100%

eof = *         ; end of file - variables below are external to saved program
capsule_x       = eof
fuel_qty        = eof+1
new_game        = eof+2
fuel_timer      = eof+3
thrust_timer    = eof+4
pause_color     = eof+5
pause_timer     = eof+6
level_number    = eof+7
lives_count     = eof+8
bonus_count     = eof+9
game_over_flag  = eof+10
exit_flag       = eof+11
menu_flag       = eof+12
ufo_y_dir       = eof+13
score           = eof+14
score_digits    = eof+16